#!/usr/bin/python
"""
Toad - To-Do List

Copyright (C) 2001 James ''Wez''Weatherall

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

ToDo:
1) Nested items
2) Context (time, network address)
3) More view mode options
4)  Rewrite database to use shelve?
    Currently, we rewrite the whole thing whenever stuff changes.
    With shelve/dbm, we would only rewrite changed stuff, I think.
    This would be better for the FLASH RAM.

"""

try: import pylau
except ImportError: pass

import gtk
try: import cPickle
except:
    import pickle
    cPickle=pickle
import sys
import time

class ToDoMgr:
    def __init__(self, filename):
        self.__setstate__({"fields":ToDo.Fields, "todos":[]})
        self.filename = filename
    def new_todo(self):
        return ToDo(self)
    def add_todo(self, todo):
        if todo not in self.list:
            self.modified = 1
            self.list.append(todo)
    def del_todo(self, todo):
        if todo in self.list:
            self.modified = 1
            self.list.remove(todo)
    def sort(self):
        self.list.sort(self.cmp_todos)
    def cmp_todos(self, a, b):
        if (((not a.complete) and b.complete) or
            ((a.complete == b.complete) and
             ((self.todo_pri(a) > self.todo_pri(b)) or
              ((self.todo_pri(a) == self.todo_pri(b)) and
               (a.text < b.text))))):
            return -1
        else:
            return 0
    def dump(self):
        if not self.timer:
            # - Start the commit timer
            self.timer = gtk.timeout_add(30000, self.on_dump_timeout)
    def dump_now(self):
        # - Do we need to dump the database?
        do_dump = self.modified
        self.modified = 0
        for todo in self.list:
            if todo.modified:
                do_dump = 1
                todo.modified = 0
        if do_dump:
            print "toad:committed database"
            cPickle.dump(self, open(self.filename, "wb"))
        return do_dump
    def on_dump_timeout(self):
        done = self.dump_now()
        if not done: self.timer = None
        return done
    def set_context(self, context):
        self.context = context
    def todo_pri(self, todo):
        pri = todo.pri
        if self.context and (self.context == todo.context):
            pri = pri+2
        return pri
    def __setstate__(self, state):
        self.context=None
        self.contexts = []
        self.timer = None
        self.modified = 0
        if type(state) == type([]):
            print "warning: loading old format"
            self.list = state
            for todo in self.list:
                todo.mgr = self
            return
        elif type(state) == type(()):
            print "warning: patching old format"
            state = {"fields":state[0],
                     "todos":state[1]}
        # - Load the standard format
        self.list=[]
        fields = state["fields"]
        for item in state["todos"]:
            todo = ToDo(self)
            for i in range(len(fields)):
                setattr(todo, fields[i], item[i])
            self.list.append(todo)
        if state.has_key("contexts"): self.contexts=state["contexts"]
                
    def __getstate__(self):
        def f(t):
            l = []
            for name in ToDo.Fields:
                l.append(getattr(t, name))
            return l
        return {"fields":ToDo.Fields,
                "todos":map(f, self.list),
                "contexts":self.contexts}

class ToDo:
    Fields = ["text", "pri", "start", "end", "context", "complete", "children"]
    def __init__(self, mgr):
        self.mgr = mgr
        self.modified = 0
        
        self.text = ""
        self.pri = 2.0
        self.start = None
        self.end = None
        self.context = None
        self.complete = 0
        self.children = []
    def dump(self):
        self.mgr.dump()
    def dump_now():
        self.mgr.dump_now()
    def __setstate__(self, state):
        self.__dict__ = state
        self.modified = 0
    def __getstate__(self):
        state = {}
        for field in ToDo.Fields:
            state[field] = getattr(self, field)
        return state

def HButtonBox(*buttons):
    box=gtk.GtkHButtonBox()
    box.set_spacing(0)
    box.set_child_ipadding_default(0, 0)
    box.set_child_size_default(0,0)
    for i in buttons:
        box.add(i)
    return box

def Button(title, func, name=None):
    b=gtk.GtkButton(title)
    if name: b.set_name(name)
    b.connect("clicked", func)
    b.show()
    return b

def Frame(title, content):
    f=gtk.GtkFrame(title)
    f.add(content)
    return f

def MenuItem(title, func, data=None, name=None):
    i=gtk.GtkMenuItem(title)
    if func: i.connect("activate", func)
    if name: i.set_name(name)
    if data: i.set_data(data[0], data[1])
    i.show()
    return i

def Menu(items):
    m=gtk.GtkMenu()
    for i in items:
        m.add(i)
    return m

class ToDoView(gtk.GtkWindow):
    def __init__(self, todo):
        gtk.GtkWindow.__init__(self)
        self.set_name("ToadTodo")
        self.todo = todo
        
        self.set_title("To-Do Item")
        self.set_default_size(240,320)

        # - Button bar
        buttons=HButtonBox(Button("Save", self.OnSaveClicked, "Commit"),
                           Button("Cancel", self.OnCancelClicked, "Discard"))

        # - To-Do item fields
        self.text = gtk.GtkText()
        self.text.set_usize(32,32)
        self.text.set_editable(1)
        self.text.insert_defaults(self.todo.text)
        self.priority = self.todo.pri
        hbox_pri = gtk.GtkHBox()
        button_pri = None
        for i in range(5):
            button_pri = gtk.GtkRadioButton(button_pri, str(i))
            button_pri.connect("toggled", self.OnSetPriority, float(i))
            if i == int(self.todo.pri): button_pri.set_active(1)
            hbox_pri.add(button_pri)
        self.context=gtk.GtkCombo()
        if self.todo.mgr: self.context.set_popdown_strings([""]+self.todo.mgr.contexts)
        for item in self.context.list.children():
            if item.children()[0].get() == "":
                item.children()[0].set_text("<None>")
                self.context.set_item_string(item, "")
        if self.todo.context: self.context.entry.set_text(self.todo.context)
        self.complete = gtk.GtkCheckButton("Completed")
        self.complete.set_active(self.todo.complete)

        vbox = gtk.GtkVBox()
        vbox.pack_start(Frame("Text", self.text))
        vbox.pack_start(Frame("Priority", hbox_pri),0,0)
##        vbox.pack_start(frame_start)
##        vbox.pack_start(frame_end)
        vbox.pack_start(Frame("Context", self.context),0,0)
        vbox.pack_start(self.complete,0,0)
        vbox.pack_end(buttons, 0, 0)
        self.add(vbox)

        self.show_all()
        self.text.grab_focus()

    def OnSetPriority(self, widget, value):
        if widget.get_active(): self.priority = value
    def OnSaveClicked(self, widget):
        if self.todo.context != self.context.entry.get_text():
            self.todo.modified = 1
            context = self.context.entry.get_text()
            if context:
                if self.todo.mgr:
                    if context in self.todo.mgr.contexts:
                        context = self.todo.mgr.contexts[self.todo.mgr.contexts.index(context)]
                    else:
                        self.todo.mgr.contexts.append(context)
            else:
                context = None
            self.todo.context = context
        if self.todo.text != self.text.get_chars(0,-1):
            self.todo.modified = 1
            self.todo.text = self.text.get_chars(0,-1)
        if self.todo.complete != self.complete.get_active():
            self.todo.modified = 1
            self.todo.complete = self.complete.get_active()
        if self.todo.pri != self.priority:
            self.todo.modified = 1
            self.todo.pri = self.priority
        
        self.todo.mgr.add_todo(self.todo)
        self.todo.dump()
        self.destroy()
    def OnCancelClicked(self, widget):
        self.destroy()

## _todo_colour = [gtk.GdkColor(255, 0, 0),
##                  gtk.GdkColor(128, 0, 0),
##                  gtk.GdkColor(128, 128, 0),
##                  gtk.GdkColor(0, 128, 0),
##                  gtk.GdkColor(0, 255, 0)]

_ToDoMgrView_Filter = {
    "All" : lambda x:1,
    "ToDo" : lambda x:not x.complete,
    "Done" : lambda x:x.complete
    }

class _ToDoMgrView_config:
    def __init__(self):
        self.view = "ToDo"
        self.context = None

class ToDoMgrView(gtk.GtkWindow):
    def __init__(self, todo_mgr):
        gtk.GtkWindow.__init__(self)
        self.set_name("ToadMgr")
        self.mgr = todo_mgr
        
        self.set_title("Toad")
        self.set_default_size(240,320)
        self.connect("destroy", self.OnDestroy)

        try: self.pconfig = cPickle.load(open("/tmp/.toad-config", "rb"))
        except: self.pconfig = _ToDoMgrView_config()

        # - Menu bar
        menu_hbox = gtk.GtkHBox()
        menu = gtk.GtkMenuBar()
        file_item = MenuItem("Toad", None)
        file_menu = gtk.GtkMenu()
        quit_item = MenuItem("Quit", self.destroy)
        file_menu.add(quit_item)
        file_item.set_submenu(file_menu)
        menu.add(file_item)

        # - View options
        view = gtk.GtkOptionMenu()
        items = []
        for viewname in _ToDoMgrView_Filter.keys():
            item = MenuItem(viewname, self.OnViewSelected, ("view", viewname))
            if viewname==self.pconfig.view: items.insert(0, item)
            else: items.append(item)
        view.set_menu(Menu(items))

        # - Contexts
        context = gtk.GtkOptionMenu()
        self.context_menu = Menu([MenuItem(self.pconfig.context or "<None>",
                                          self.OnContextSelected,
                                          ("context", self.pconfig.context))])
        context.set_menu(self.context_menu)
        self.mgr.set_context(self.pconfig.context)

        menu_hbox.pack_start(menu)
        menu_hbox.pack_end(context)
        menu_hbox.pack_end(view)

        # - Button bar
        buttons = HButtonBox(Button("New", self.OnNewClicked, "Create"),
                             Button("Edit", self.OnEditClicked),
                             Button("Destroy", self.OnDestroyClicked, "Destroy"),
                             Button("Quit", self.destroy))

        # - Create the list
        self.list = gtk.GtkList()
        self.list.connect("select-child", self.OnToDoSelected)
        self.selected = None
        self.Refresh()

        # - Build the window layout
        vbox = gtk.GtkVBox()
        vbox.pack_start(menu_hbox, 0, 0)
        list_wnd = gtk.GtkScrolledWindow()
        list_wnd.add_with_viewport(self.list)
        list_wnd.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        vbox.add(list_wnd)
        vbox.pack_end(buttons, 0, 0)
        self.add(vbox)

        self.show_all()

    def OnDestroy(self, widget):
        self.mgr.dump_now()
        try: cPickle.dump(self.pconfig, open("/tmp/.toad-config", "wb"))
        except: print "warning: unable to save view mode"
        gtk.mainquit()

    # - Set the view mode
    def OnViewSelected(self, widget):
        self.pconfig.view = widget.get_data("view")
        self.Refresh()

    # - Context
    def OnContextSelected(self, widget):
        self.mgr.set_context(widget.get_data("context"))
        self.pconfig.context=widget.get_data("context")
        self.Refresh()

    # - The bottom Button Bar
    def OnNewClicked(self, widget):
        todo = self.mgr.new_todo()
        todo_view = ToDoView(todo)
        todo_view.connect("destroy", self.OnToDoEditClosed, todo_view)
    def OnEditClicked(self, widget):
        if self.selected:
            todo_view = ToDoView(self.selected)
            todo_view.connect("destroy", self.OnToDoEditClosed, todo_view)
    def OnDestroyClicked(self, widget):
        if self.selected:
            self.mgr.del_todo(self.selected)
            self.selected = None
            self.Refresh()

    # - ToDo item actions
    def OnToDoChecked(self, widget, todo):
        todo.complete = widget.get_active()
        todo.modified = 1
        self.Refresh()
    def OnToDoEditClosed(self, widget, view):
        todo = view.todo
        self.selected = todo
        self.Refresh(todo)
    def OnToDoSelected(self, widget, child):
        self.selected = child.get_data("todo")

    # - Redrawing the list
    def Refresh(self, todo_refresh=None):
        # - Sort and filter To-Dos
        self.mgr.sort()
        todos = filter(_ToDoMgrView_Filter[self.pconfig.view], self.mgr.list)

        # - Remove dead display entries
        for item in self.list.children()[:]:
            item.destroy()
        shown = []

        context_shown=[]
        for item in self.context_menu.children()[:]:
            context = item.get_data("context")
            if context and (context not in self.mgr.contexts):
                item.destroy()
            else:
                context_shown.append(item.get_data("context"))

        # - Insert new items
        items = []
        select = None
        for todo in todos:
            if todo not in shown:
                item = gtk.GtkListItem()
                item.set_data("todo", todo)
                hbox = gtk.GtkHBox()
                check = gtk.GtkCheckButton()
                check.set_active(todo.complete)
                check.connect("clicked", self.OnToDoChecked, todo)
                text = gtk.GtkLabel(todo.text)
                text.set_line_wrap(1)
                hbox.pack_start(check, 0, 0)
                hbox.pack_start(text)
                hbox.pack_start(gtk.GtkLabel(str(todo.pri)), 0, 0)
                item.add(hbox)
                item.show_all()
                if todo == self.selected: select = item
                items.append(item)
        self.list.append_items(items)
        if select: select.select()
        self.mgr.dump()

        for context in self.mgr.contexts+[None]:
            if context not in context_shown:
                self.context_menu.add(MenuItem(context or "<None>",
                                               self.OnContextSelected,
                                               ("context", context)))

def Main(argv):
    if len(argv) >= 2:
        filename = argv[1]
    else:
        filename = "/tmp/toad.pickle"
    try:
        file = open(filename, "rb")
        todo_mgr = cPickle.load(file)
        todo_mgr.filename = filename
    except:
        todo_mgr = ToDoMgr(filename)
    todo = ToDoMgrView(todo_mgr)
    gtk.mainloop()

if __name__ == "__main__": Main(sys.argv)
